import warnings
from types import ModuleType
from typing import Any
from typing import Final

from django.conf import settings
from django.core.signals import setting_changed
from django.dispatch import receiver
from django.utils.module_loading import import_string

from meringue.conf import default_settings


SETTING_KEY: Final[str] = "MERINGUE"
"""
Parameter name in django settings for meringue settings.

Examples:
    ```py title="settings.py"
    MERINGUE = {
        "FRONTEND_URL": "http://meringue.local:9000/",
    }
    ```
"""


DEPRECATED_PARAMS: Final[dict[str, str]] = {}
"""
Dict with deprecated options and warning texts for them.

Examples:
    ```py
    DEPRECATED_PARAMS = {
        "PROTOCOL": "The `PROTOCOL` option is deprecated, use `BACKEND_PROTOCOL` instead.",
    }
    ```
"""

PARAMS_TO_IMPORT: Final[list[str]] = [
    "UPLOAD_RENAME_HANDLER",
    "THUMBNAIL_GENERATOR_CLASS",
    "THUMBNAIL_STORAGE_GETTER",
    "THUMBNAIL_IMAGE_CLASS",
    "THUMBNAIL_PROPERTIES",
    "THUMBNAIL_ACTIONS",
]
"""
List of options that contain the path to the module and must be imported.

Examples:
    ```py
    PARAMS_TO_IMPORT = [
        "UPLOAD_RENAME_HANDLER",
    ]
    ```
"""


def import_list(path_list: list[str]) -> list[Any]:
    """
    Import a list of dotted modules.

    Attributes:
        path_list: List of dotted paths.

    Raises:
        TypeError: Wrong dotted path data type for import.
        ImportError: Import error.

    Returns:
        List of imported attributes/classes.
    """

    result = []

    for dotted_path in path_list:
        if not isinstance(dotted_path, str):
            msg = f"`{type(dotted_path).__name__}` data type is not supported for import."
            raise TypeError(msg, dotted_path)

        try:
            result.append(import_string(dotted_path))
        except ImportError as err:
            msg = f"`{dotted_path}` import error."
            raise ImportError(msg, path=dotted_path) from err

    return result


def import_dict(path_dict: dict[str, str]) -> dict[str, Any]:
    """
    Import a dictionary with dotted path as key value.

    Attributes:
        path_dict: Dictionary with dotted path as key value.

    Raises:
        TypeError: Wrong dotted path data type for import.
        ImportError: Import error.

    Returns:
        Dict with imported attributes/classes as key value.
    """

    result = {}

    for key, dotted_path in path_dict.items():
        if not isinstance(dotted_path, str):
            msg = f"`{type(dotted_path).__name__}` data type is not supported for import."
            raise TypeError(msg, dotted_path)

        try:
            result[key] = import_string(dotted_path)
        except ImportError as err:
            msg = f"`{dotted_path}` import error."
            raise ImportError(msg, path=dotted_path) from err

    return result


def import_parameter(value: str | list[str] | dict[str, str]) -> Any | list[Any] | dict[str, Any]:
    """
    Method for importing modules in settings parameter.

    Attributes:
        value: Settings parameter value to import.

    Returns:
        Imported parameter.
    """

    if isinstance(value, str):
        try:
            return import_string(value)
        except ImportError as err:
            msg = f"`{value}` import error."
            raise ImportError(msg, path=value) from err

    if isinstance(value, list):
        return import_list(value)

    if isinstance(value, dict):
        return import_dict(value)

    msg = f"`{type(value).__name__}` data type is not supported for import."
    raise TypeError(msg, value)


class Settings:
    """
    A settings object.

    The settings are obtained from the django settings by the name of the key which should contain
    an object with all the application settings.
    """

    def __init__(
        self,
        setting_key: str,
        defaults: dict[str, str],
        deprecated_params: dict[str, str] | None = None,
        params_to_impoprt: list[str] | None = None,
    ):
        """
        Attributes:
            setting_key: Settings key in django settings list.
            defaults: Dict with default parameter values. Used as a list of available settings.
            deprecated_params: Dict with deprecated options and warning texts for them.
            params_to_impoprt: List of options that contain the path to the module and must be
                imported.
        """

        self.setting_key = setting_key
        if isinstance(defaults, ModuleType):
            self.defaults = {}
            for key in dir(defaults):
                if key.isupper():
                    self.defaults[key] = getattr(defaults, key)
        else:
            self.defaults = defaults
        self.deprecated_params = deprecated_params or {}
        self.params_to_impoprt = params_to_impoprt or []
        self._cached_attrs = set()
        self.reset()

    def __getattr__(self, attr: str) -> Any:
        """
        Gets the parameter value and caches it in the attributes of the settings object.

        Attributes:
            attr: Setting parameter name.

        Raises:
            AttributeError: Error when trying to get an unregistered parameter.

        Warns:
            DeprecationWarning: A warning that the parameter is deprecated.

        Returns:
            Setting value.
        """

        if attr not in self.defaults:
            raise AttributeError("Invalid setting key: '%s'" % attr)

        if attr in self.deprecated_params:
            warnings.warn(self.deprecated_params[attr], DeprecationWarning, stacklevel=2)

        val = self.user_params.get(attr, self.defaults[attr])

        if attr in self.params_to_impoprt:
            try:
                val = import_parameter(val)

            except ImportError as err:
                msg = f"Error importing `{err.path}` attribute/class in `{attr}` parameter."
                raise ImportError(msg) from err

            except TypeError as err:
                msg = (
                    f"The `{err.args[1]}` value of the `{attr}` parameter is not available for "
                    "import."
                )
                raise TypeError(msg) from None

        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reset(self):
        """
        Reset downloaded settings, as well as clearing the cache.
        """

        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs = set()
        self.user_params = getattr(settings, self.setting_key, {})


m_settings = Settings(SETTING_KEY, default_settings, DEPRECATED_PARAMS, PARAMS_TO_IMPORT)


@receiver(setting_changed)
def reset_settings(*args, **kwargs):
    """
    Settings change signal handler.
    """

    if kwargs["setting"] == SETTING_KEY:
        m_settings.reset()



import mimetypes
from pathlib import Path
from urllib.parse import quote
from urllib.parse import urljoin

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.http import FileResponse
from django.http import Http404
from django.http import HttpResponse

from meringue.conf import m_settings
from meringue.protected.fields import ProtectedFieldFile
from meringue.protected.fields import ProtectedImageFieldFile


def protected_file_view(request, contenttype_id, field, pk):
    """
    В продакшен моде вьха редиректит на урл файла, где нжинкс отдаёт этот файл
    Защита работает за счёт параметра internal
    https://nginx.org/en/docs/http/ngx_http_core_module.html#internal

    ```conf
    location /media/protected/ {
        internal;
        alias /home/rpl/rpl-2023-back/public/media/protected/;
    }
    ```
    """

    contenttype = ContentType.objects.get(id=contenttype_id)

    if not request.user.has_perm(f"{contenttype.app_label}.view_{contenttype.model}"):
        raise PermissionDenied()

    model = contenttype.model_class().objects.get(pk=pk)
    file = getattr(model, field)

    if not file:
        raise Http404()

    if m_settings.PROTECTED_SERVE_WITH_NGINX:
        file_name = Path(file.name).name
        response = HttpResponse()
        response["Content-Type"] = mimetypes.guess_type(file.path)[0]
        response["Content-Disposition"] = f"inline; filename={quote(file_name)}"

        if isinstance(file, ProtectedFieldFile | ProtectedImageFieldFile):
            redirect_url = file.original_url
        else:
            redirect_url = file.url

        response["X-Accel-Redirect"] = redirect_url
        return response

    return FileResponse(open(file.path, "rb"))



import copy
import logging
from pathlib import Path
from typing import Any
from typing import Final

from django.core.files.storage import Storage
from django.utils.functional import lazy

from PIL import Image

from meringue.conf import m_settings
from meringue.thumbnail import constants
from meringue.thumbnail.exceptions import WrongActionOrPropertyError
from meringue.thumbnail.exceptions import WrongFormatError
from meringue.thumbnail.images import DefaultThumbnailImage
from meringue.thumbnail.storage import default_storage
from meringue.thumbnail.types import JobChainType


logger = logging.getLogger("meringue.thumbnail")


DEFAULT_OPTIONS: Final[dict[str, Any]] = {
    constants.PROP_BG_COLOR: m_settings.THUMBNAIL_DEFAULT_BG_COLOR,
    constants.PROP_CROP_METHOD: m_settings.THUMBNAIL_DEFAULT_CROP_METHOD,
    constants.PROP_MAX_HEIGHT: None,
    constants.PROP_MAX_WIDTH: None,
    constants.PROP_RESIZE_METHOD: m_settings.THUMBNAIL_DEFAULT_RESIZE_METHOD,
    constants.PROP_RESIZE_STRATEGY: m_settings.THUMBNAIL_DEFAULT_RESIZE_STRATEGY,
}


class Thumbnailer:
    """
    Thumbnail generator class.

    Todo:

    * aliases for job changes
    """

    def __init__(
        self,
        image_path: Path,
        job_chain: JobChainType,
        storage: Storage | None = None,
    ):
        """
        Attributes:
            image_path: Source image path.
            job_chain: Thumbnail job chain.
            storage: File storage.
        """

        if not image_path.is_absolute():
            msg = "`image_path` must be absolute"
            raise AttributeError(msg, name="image_path")

        self.source_image_path = image_path
        self.job_chain = job_chain
        self.storage = storage or default_storage

    def make_thumbnail(self) -> Image:
        """
        Makes a thumbnail based on a quest chain.

        Raises:
            WrongActionOrPropertyError: Invalid action or property.

        Returns:
            Pillow image.
        """

        logger.debug(f"Thumbnail is made using job chain: {', '.join(self.job_chain)}")

        # openingn and prepearing file
        image = Image.open(self.source_image_path)
        image = image.convert("RGBA")

        # prepearing options
        options = copy.deepcopy(DEFAULT_OPTIONS)
        options[constants.PROP_CURRENT_SIZE] = [float(i) for i in image.size]
        options[constants.PROP_NEW_SIZE] = options[constants.PROP_CURRENT_SIZE].copy()

        # applying modifiers
        for job in self.job_chain:
            if ":" in job:
                # property
                prop, args = job.split(":")
                options.update(m_settings.THUMBNAIL_PROPERTIES[prop](args, options))

            elif job in m_settings.THUMBNAIL_ACTIONS:
                # action
                image = m_settings.THUMBNAIL_ACTIONS[job](image, options)

            else:
                raise WrongActionOrPropertyError(job)

        return image

    def get_thumbnail(self, out_format: str, **kwargs) -> DefaultThumbnailImage:
        """
        Generates and returns a preview image in the specified format.

        Attributes:
            out_format: Output image format.
            **kwargs: Output image save extra params.

        Raises:
            WrongFormatError: Format is not supported.

        Returns:
            Thumbnail image.
        """

        if out_format not in constants.EXTENSIONS_BY_FORMATS:
            raise WrongFormatError(out_format)

        image = lazy(self.make_thumbnail, Image.Image)()
        thumbnail_image = DefaultThumbnailImage(
            image_path=self.source_image_path,
            job_chain=self.job_chain,
            out_format=out_format,
            image=image,
            storage=self.storage,
        )
        thumbnail_image.save(**kwargs)

        return thumbnail_image


DefaultThumbnailer: Final = m_settings.THUMBNAIL_GENERATOR_CLASS



import re

from meringue.thumbnail.constants import CROP_METHOD_BOTTOM
from meringue.thumbnail.constants import CROP_METHOD_CENTER
from meringue.thumbnail.constants import CROP_METHOD_LEFT
from meringue.thumbnail.constants import CROP_METHOD_RIGTH
from meringue.thumbnail.constants import CROP_METHOD_TOP
from meringue.thumbnail.constants import PROP_BG_COLOR
from meringue.thumbnail.constants import PROP_CROP_METHOD
from meringue.thumbnail.constants import PROP_CURRENT_SIZE
from meringue.thumbnail.constants import PROP_MAX_HEIGHT
from meringue.thumbnail.constants import PROP_MAX_WIDTH
from meringue.thumbnail.constants import PROP_NEW_SIZE
from meringue.thumbnail.constants import PROP_RESIZE_METHOD
from meringue.thumbnail.constants import PROP_RESIZE_STRATEGY
from meringue.thumbnail.constants import RESIZE_METHOD_LIST
from meringue.thumbnail.constants import RESIZE_STRATEGY_LIST
from meringue.thumbnail.exceptions import WrongPropertyOptionError


# cm - crop method
def set_crop_method(value, opt):
    methods = value.split(" ")

    if len(methods) == 1:
        if methods[0] != CROP_METHOD_CENTER:
            msg = "For property `{property}` if it contains one value it can only be `{option}`"
            raise WrongPropertyOptionError(PROP_CROP_METHOD, CROP_METHOD_CENTER, msg)

        methods = [CROP_METHOD_CENTER, CROP_METHOD_CENTER]

    elif len(methods) == 2:  # noqa: PLR2004
        if methods[0] not in [CROP_METHOD_RIGTH, CROP_METHOD_LEFT, CROP_METHOD_CENTER]:
            msg = "Value `{option}` is invalid for property `{property}` at the first position."
            raise WrongPropertyOptionError(PROP_CROP_METHOD, methods[0], msg)

        if methods[1] not in [CROP_METHOD_TOP, CROP_METHOD_BOTTOM, CROP_METHOD_CENTER]:
            msg = "Value `{option}` is invalid for property `{property}` at the second position."
            raise WrongPropertyOptionError(PROP_CROP_METHOD, methods[1], msg)

    else:
        msg = "The property `{property}` must have one or two options."
        raise WrongPropertyOptionError(PROP_CROP_METHOD, None, msg)

    return {PROP_CROP_METHOD: methods}


# rm - resize method
def set_resize_method(method, opt):
    if method not in RESIZE_METHOD_LIST:
        raise WrongPropertyOptionError(PROP_RESIZE_METHOD, method)

    return {PROP_RESIZE_METHOD: method}


# rs - resize strategy
def set_resize_strategy(strategy, opt):
    if strategy not in RESIZE_STRATEGY_LIST:
        raise WrongPropertyOptionError(PROP_RESIZE_STRATEGY, strategy)

    return {PROP_RESIZE_STRATEGY: strategy}


# s - size
def set_size(raw, opt):
    size_re = re.compile(r"(\d+)?x(\d+)?")
    size = [float(i or 0) for i in size_re.search(raw).groups()]

    # расчитываем относительные размеры если один из размеров не указан
    size[0] = (
        size[0] if size[0] else size[1] * opt[PROP_CURRENT_SIZE][0] / opt[PROP_CURRENT_SIZE][1]
    )
    size[1] = (
        size[1] if size[1] else size[0] * opt[PROP_CURRENT_SIZE][1] / opt[PROP_CURRENT_SIZE][0]
    )

    return {PROP_NEW_SIZE: size}


# maxw - max width
def set_max_width(width, opt):
    return {PROP_MAX_WIDTH: float(width)}


# maxh - max height
def set_max_height(height, opt):
    return {PROP_MAX_HEIGHT: float(height)}


# c - background color
def set_bg_color(color, opt):
    tmp = ()

    for i in color.split(" "):
        tmp += (int(i),)

    return {PROP_BG_COLOR: tmp}



__version__ = "1.2.0.dev5"
"""
To update the version, use [hatch version](https://hatch.pypa.io/latest/version/#updating)

show current version:

```console
$ hatch version
1.0.0a0
```

update patch and beta version from `1.0.0a0` to `1.0.1b0`

```console
$ hatch version patch,beta
Old: 1.0.0a0
New: 1.0.1b0
```
"""



from django.core.exceptions import PermissionDenied
from django.http.response import Http404

from rest_framework import exceptions
from rest_framework import views

from meringue.api.utils import render_error_details

try:
    from rest_framework_simplejwt.exceptions import DetailDictMixin
except ImportError:
    DetailDictMixin = None


def exception_handler(exc, context):
    """
    Error handler returning message and error code pairs.

    The handler is a wrapper over the standard handler `rest_framework.views.exception_handler`.
    """

    response = views.exception_handler(exc, context)

    if response is None:
        return response

    if isinstance(exc, (Http404, PermissionDenied)):
        # django Http404 and PermissionDenied errors are substituted for drf errors,
        # in `rest_framework.views.exception_handler` method.
        response.data = render_error_details(response.data["detail"])

    elif DetailDictMixin and isinstance(exc, DetailDictMixin):
        response.data = render_error_details(exc)

    elif isinstance(exc, exceptions.APIException):
        response.data = render_error_details(exc.detail)

    return response



import datetime as dt

from django import template
from django.utils import timezone
from django.utils.safestring import mark_safe

from meringue.conf import m_settings
from meringue.core.utils import format_date_from_to


register = template.Library()


@register.simple_tag
def cop_year() -> str:
    """
    A tag that displays the year or range of years for the copyright string in `YYYY-YYYY` format.

    Examples:
        ```jinja
        <p>Copyright © {% cop_year %} My company</p>
        ```

    Raises:
        Exception: To use the `cop_year` tag, you must fill in the `COP_YEAR` parameter in the
            meringue settings

    Returns:
        Year for copyrights.
    """

    if m_settings.COP_YEAR is None:
        msg = (
            "To use the `cop_year` tag, you must fill in the `COP_YEAR` parameter in the "
            "meringue settings"
        )
        raise Exception(msg)

    year = timezone.localtime().year

    if year == m_settings.COP_YEAR or year - m_settings.COP_YEAR < m_settings.COP_YEARS_DIFF:
        return year

    return mark_safe(f"{m_settings.COP_YEAR}&mdash;{year}")  # noqa: S308


@register.simple_tag
def date_range(date_start: dt.date, date_end: dt.date) -> str:
    """
    Return range of date in one of the following formats:

    * `DD.MM.YYYY - DD.MM.YYYY`
    * `DD.MM - DD.MM.YYYY`
    * `DD - DD.MM.YYYY`
    * `DD.MM.YYYY`

    Attributes:
        date_start: Period start date.
        date_end: Period end date.

    Examples:
        ```jinja
        {% date_range date_start date_end %}
        ```

    Returns:
        Date period.
    """

    tmp_result = format_date_from_to(date_start, date_end, "-")

    if date_start == date_end:
        return tmp_result

    tmp_date_start, tmp_date_end = tmp_result.split(" - ")

    return mark_safe(f"{tmp_date_start}&nbsp;&mdash; {tmp_date_end}")  # noqa: S308



from typing import Final


# Properties keys
PROP_BG_COLOR: Final[str] = "bg_color"
PROP_CROP_METHOD: Final[str] = "crop_method"
PROP_CURRENT_SIZE: Final[str] = "current_size"
PROP_MAX_HEIGHT: Final[str] = "max_height"
PROP_MAX_WIDTH: Final[str] = "max_width"
PROP_NEW_SIZE: Final[str] = "new_size"
PROP_RESIZE_METHOD: Final[str] = "resize_method"
PROP_RESIZE_STRATEGY: Final[str] = "resize_strategy"


# Crop methods
CROP_METHOD_CENTER: Final[str] = "center"
CROP_METHOD_TOP: Final[str] = "top"
CROP_METHOD_RIGTH: Final[str] = "rigth"
CROP_METHOD_BOTTOM: Final[str] = "bottom"
CROP_METHOD_LEFT: Final[str] = "left"
CROP_METHOD_LIST: Final[list[str]] = [
    CROP_METHOD_CENTER,
    CROP_METHOD_TOP,
    CROP_METHOD_RIGTH,
    CROP_METHOD_BOTTOM,
    CROP_METHOD_LEFT,
]

# Resize methods
RESIZE_METHOD_COVER: Final[str] = "cover"
RESIZE_METHOD_CONTAIN: Final[str] = "contain"
RESIZE_METHOD_STRETCH: Final[str] = "stretch"
RESIZE_METHOD_LIST: Final[list[str]] = [
    RESIZE_METHOD_COVER,
    RESIZE_METHOD_CONTAIN,
    RESIZE_METHOD_STRETCH,
]

# Resize strategy
RESIZE_STRATEGY_DO_NOT_INCREASE_SIZE: Final[str] = "no_increase"
RESIZE_STRATEGY_STANDART: Final[str] = "standart"
RESIZE_STRATEGY_DO_NOT_REDUCE_SIZE: Final[str] = "no_reduce"
RESIZE_STRATEGY_LIST: Final[list[str]] = [
    RESIZE_STRATEGY_DO_NOT_INCREASE_SIZE,
    RESIZE_STRATEGY_STANDART,
    RESIZE_STRATEGY_DO_NOT_REDUCE_SIZE,
]

# Formats
FORMAT_BMP: Final[str] = "BMP"
"""Format BMP"""
FORMAT_EPS: Final[str] = "EPS"
"""Format EPS"""
FORMAT_GIF: Final[str] = "GIF"
"""Format GIF"""
FORMAT_ICO: Final[str] = "ICO"
"""Format ICO"""
FORMAT_JPEG: Final[str] = "JPEG"
"""Format JPEG"""
FORMAT_PNG: Final[str] = "PNG"
"""Format PNG"""
FORMAT_TIFF: Final[str] = "TIFF"
"""Format TIFF"""
FORMAT_WEBP: Final[str] = "WEBP"
"""Format WEBP"""

# Extension
EXTENSION_BMP: Final[str] = ".bmp"
"""Extension .bmp"""
EXTENSION_EPS: Final[str] = ".eps"
"""Extension .eps"""
EXTENSION_GIF: Final[str] = ".gif"
"""Extension .gif"""
EXTENSION_ICO: Final[str] = ".ico"
"""Extension .ico"""
EXTENSION_JPG: Final[str] = ".jpg"
"""Extension .jpg"""
EXTENSION_JPEG: Final[str] = ".jpeg"
"""Extension .jpeg"""
EXTENSION_JFIF: Final[str] = ".jfif"
"""Extension .jfif"""
EXTENSION_PNG: Final[str] = ".png"
"""Extension .png"""
EXTENSION_TIFF: Final[str] = ".tiff"
"""Extension .tiff"""
EXTENSION_WEBP: Final[str] = ".webp"
"""Extension .webp"""


FORMATS_BY_EXTENSIONS: Final[dict[str, str]] = {
    EXTENSION_BMP: FORMAT_BMP,
    EXTENSION_EPS: FORMAT_EPS,
    EXTENSION_GIF: FORMAT_GIF,
    EXTENSION_ICO: FORMAT_ICO,
    EXTENSION_JPG: FORMAT_JPEG,
    EXTENSION_JPEG: FORMAT_JPEG,
    EXTENSION_JFIF: FORMAT_JPEG,
    EXTENSION_PNG: FORMAT_PNG,
    EXTENSION_TIFF: FORMAT_TIFF,
    EXTENSION_WEBP: FORMAT_WEBP,
}
"""Formats by extensions"""


EXTENSIONS_BY_FORMATS: Final[dict[str, str]] = {
    # DIB
    # BUFR
    # PCX
    # FITS
    # GRIB
    # HDF5
    # JPEG2000
    # IM
    # MPO
    # MSP
    # PALM
    # PDF
    # PPM
    # SGI
    # SPIDER
    # TGA
    # WMF
    # XBM
    FORMAT_BMP: EXTENSION_BMP,
    FORMAT_EPS: EXTENSION_EPS,
    FORMAT_GIF: EXTENSION_GIF,
    FORMAT_ICO: EXTENSION_ICO,
    FORMAT_JPEG: EXTENSION_JPG,
    FORMAT_PNG: EXTENSION_PNG,
    FORMAT_TIFF: EXTENSION_TIFF,
    FORMAT_WEBP: EXTENSION_WEBP,
}
"""Extensions by formats"""

FORMATS_WITH_ALPHA_SUPPORT: Final[list[str]] = [
    FORMAT_BMP,
    FORMAT_EPS,
    FORMAT_GIF,
    FORMAT_ICO,
    FORMAT_PNG,
    FORMAT_TIFF,
    FORMAT_WEBP,
]
"""Formats with alpha support"""



from collections import OrderedDict
from collections import namedtuple

from rest_framework.routers import DefaultRouter

from meringue.conf import m_settings


ObjectRoute = namedtuple("ObjectRoute", ["url", "mapping", "name", "detail", "initkwargs"])
"""
Object predefined detail route.
"""


class MeringueRouter(DefaultRouter):
    """
    The default router is inherited from the original default router but with some modifications.
    """

    root_view_name = "index"
    routes = [  # noqa: RUF012
        # Object predefined detail route.
        ObjectRoute(
            url=r"^{prefix}{trailing_slash}$",
            mapping={
                "get": "retrieve",
                "put": "update",
                "patch": "partial_update",
                "delete": "destroy",
            },
            name="{basename}",
            detail=True,
            initkwargs={},
        ),
        *DefaultRouter.routes,
    ]

    @property
    def include_root_view(self):
        return m_settings.API_ENABLE_ROOT_VIEW

    def get_api_root_view(self, api_urls=None):
        """
        Return a basic root view.
        """

        api_root_dict = OrderedDict()
        object_name = self.routes[0].name
        list_name = self.routes[1].name
        for prefix, viewset, basename in self.registry:
            if getattr(viewset, "m_object_detail", False):
                api_root_dict[prefix] = object_name.format(basename=basename)
            else:
                api_root_dict[prefix] = list_name.format(basename=basename)

        return self.APIRootView.as_view(api_root_dict=api_root_dict)

    def get_routes(self, viewset):
        """
        Extra routes are filtered according to the new route type.
        """

        tmp_routes = super().get_routes(viewset)
        is_object_viewset = getattr(viewset, "m_object_detail", False)

        routes = []
        for route in tmp_routes:
            is_object_route = isinstance(route, ObjectRoute)

            if is_object_viewset:
                is_list_view = route.name == "{basename}-list"
                is_detail_view = route.name == "{basename}-detail"

                if is_list_view or is_detail_view:
                    # ignore default list and detail routes
                    continue

                routes.append(route)

            elif not is_object_route:
                routes.append(route)

        return routes



from django.template import Library

from meringue.thumbnail.shortcuts import get_thumbnail


register = Library()


@register.filter
def thumbnail(filename, args=""):
    """
        Фильтр обработки изображений, применяется к адресу изображения (путь до файла) в аргументах
    ожидает параметры изменения изображения разделённые запятой, последовательность действий
    соответствует последовательности введённых аргументов.
    Например:
        * последовательность s:600x400,resize,s:400x400,crop - сначала задас целевой размер
    изображения 600x400, далее  скукожит (или растянет изображение в зависимости от исходного и
    других параметров) до 600x400, потом установит новый целевой размер 400x400 и последним этапом
    будет обезание изображения в размер 400x400.
        * последовательность crop,s:400x400 не даст никакого результата, потому что новый размер
    указывается после события кропа (если параметры поменять местами, то изображение обрежется до
    квадрата 400x400).

    Аргументы фильтра:
        s:<width>x<height> - указывает целевой размер изображения для
            последующих функций
        maxw:<width> - указывает максимальную ширину
        maxh:<height> - указывает максимальную высоту
        crop - изменения размера холста до последнего установленного
        resize - изменение размера изображения до последнего установленного
        c:<color> - цвет заливки для кропа в формате rgba
            (c:255 255 255 255)
        cm:left|center|rigth top|bottom - точка отсчёта для кропа
        rm:cover|contain|stretch - метод ресайза вписать в размер или
            растянуть
        rs:no_increase|standart|no_reduce - стратегия ресайза - не давать увеличивать исходное
            изображение или уменьшать
    """

    task_list = args.split(",")
    return get_thumbnail(filename=filename, task_list=task_list)



from typing import Final

import django
from django.conf import settings
from django.core.files.storage import Storage

from meringue.conf import m_settings


def get_storage() -> Storage:
    """
    Finds the default storage class and returns its instance for use with thumbnails.

    !!! warning
        This method has not been tested with custom stores and complex parameter sets.

    Returns:
        Storage instance.
    """

    if django.VERSION < (4, 2):
        DefaultStorage = settings.DEFAULT_FILE_STORAGE  # noqa: N806
        storage = DefaultStorage(
            location=m_settings.THUMBNAIL_DIR,
            base_url=m_settings.THUMBNAIL_URL,
        )
    else:
        from django.conf import DEFAULT_STORAGE_ALIAS
        from django.core.files.storage import storages

        params = storages.backends[DEFAULT_STORAGE_ALIAS]
        params["location"] = m_settings.THUMBNAIL_DIR
        params["base_url"] = m_settings.THUMBNAIL_URL
        storage = storages.create_storage(params)

    return storage


default_storage: Final[Storage] = m_settings.THUMBNAIL_STORAGE_GETTER()



import uuid
from pathlib import Path

from django.core.files import uploadhandler

from meringue.conf import m_settings


def rename_handler(file_name: str) -> str:
    """
    Default method for renaming files when using load handlers
    [MemoryFileUploadHandler][meringue.core.upload_handlers.MemoryFileUploadHandler] and / or
    [TemporaryFileUploadHandler][meringue.core.upload_handlers.TemporaryFileUploadHandler].

    Attributes:
        file_name: Original file name.

    Returns:
        New file name.
    """
    return str(Path(file_name).with_stem(str(uuid.uuid4())))


class MemoryFileUploadHandler(uploadhandler.MemoryFileUploadHandler):
    """
    File upload handler to stream uploads into memory (used for small files).

    Examples:
        ```py title="settings.py"
        FILE_UPLOAD_HANDLERS = (
            'meringue.core.upload_handlers.MemoryFileUploadHandler',
            'django.core.files.uploadhandler.TemporaryFileUploadHandler',
        )
        ```
    """

    def new_file(
        self,
        field_name,
        file_name,
        content_type,
        content_length,
        charset=None,
        content_type_extra=None,
    ):
        new_name = m_settings.UPLOAD_RENAME_HANDLER(file_name)
        super().new_file(
            field_name,
            new_name,
            content_type,
            content_length,
            charset,
            content_type_extra,
        )


class TemporaryFileUploadHandler(uploadhandler.TemporaryFileUploadHandler):
    """
    Upload handler that streams data into a temporary file.

    Examples:
        ```py title="settings.py"
        FILE_UPLOAD_HANDLERS = (
            'django.core.files.uploadhandler.MemoryFileUploadHandler',
            'meringue.core.upload_handlers.TemporaryFileUploadHandler',
        )
        ```
    """

    def new_file(
        self,
        field_name,
        file_name,
        content_type,
        content_length,
        charset=None,
        content_type_extra=None,
    ):
        new_name = m_settings.UPLOAD_RENAME_HANDLER(file_name)
        super().new_file(
            field_name,
            new_name,
            content_type,
            content_length,
            charset,
            content_type_extra,
        )



import logging
import os
from pathlib import PurePath

from rest_framework.fields import ImageField

from meringue.thumbnail.constants import FORMATS_BY_EXTENSIONS
from meringue.thumbnail.generators import DefaultThumbnailer
from meringue.thumbnail.shortcuts import _dummyimage


logger = logging.getLogger("meringue.thumbnail")


def get_format_from_path(path):
    return FORMATS_BY_EXTENSIONS[PurePath(path).suffix.lower()]


class BaseImageField(ImageField):
    def __init__(self, size=None, job_chain=None, **kwargs):
        if bool(size) and bool(job_chain):
            msg = "Need to set `size` or `job_chain` attribute (not both)."
            raise Exception(msg)

        elif size:
            self.job_chain = [
                f"s:{size[0]}x{size[1]}",
                "rm:cover",
                "rs:no_increase",
                "resize",
            ]

        elif job_chain:
            self.job_chain = job_chain

        else:
            self.job_chain = []

        super().__init__(**kwargs)


class MImageField(BaseImageField):
    """
    Optimize image to represent them
    """

    def to_representation(self, value):
        """
        TODO:
            * Add params to get_image for optimize images
        """

        if not value:
            return None

        if os.path.isfile(value.path):
            thumbnail = DefaultThumbnailer(value.path, job_chain=self.job_chain)
            optimized_image_url = thumbnail.get_image(get_format_from_path(value.path)).url
            result = optimized_image_url

        else:
            logger.error(f"File `{value.path}` not found")
            result = _dummyimage([])

        return result


class MImageSetField(ImageField):
    """
    Image field - make multiple files in different formats.

    Can set size (eg '100x100'), job_chain (arrays of jobs by dimensions) or nothing to optimize
    images only
    """

    def __init__(
        self,
        size=None,
        dimensions=None,
        base_job_chain=None,
        job_chains=None,
        **kwargs,
    ):
        if bool(size and dimensions) and bool(job_chains):
            msg = "Need to set `size` and `dimensions` or `job_chains` attribute (not both)."
            raise Exception(msg)

        elif size:
            self.job_chains = {}

            if 1 not in dimensions:
                dimensions.insert(0, 1)

            for dimension in dimensions:
                self.job_chains[dimension] = [
                    f"s:{round(size[0]*dimension)}x{round(size[1]*dimension)}",
                    *(base_job_chain or ["rm:cover", "resize", "crop"]),
                ]

        elif job_chains:
            self.job_chains = job_chains

        else:
            self.job_chains = {}

        self.spactacular_annotate()
        super().__init__(**kwargs)

    def spactacular_annotate(self):
        items = {
            "type": "object",
            "properties": {
                "srcset": {
                    "type": "string",
                    "example": (
                        "https://www.fillmurray.com/256/256 1x, "
                        "https://www.fillmurray.com/512/512 2x",
                    ),
                },
                "type": {
                    "title": "Image MIME type",
                    "type": "string",
                    "format": "MIME",
                    "example": "image/webp",
                },
            },
        }

        self._spectacular_annotation = {
            "field": {
                "type": "array",
                "items": items,
            },
            "field_component_name": None,
        }

    def to_representation(self, value):
        """
        Creates multiple images. Image with format as original image and webp.

        TODO:
            * Add params to get_image for optimize images
            * In future make av1 image
            * Check enabled webp
        """

        if not value:
            return None

        if not os.path.isfile(value.path):
            logger.error(f"File `{value.path}` not found")
            return [
                {
                    "url": _dummyimage(self.job_chains[1]),
                    "type": "image/png",
                },
            ]

        original_format = get_format_from_path(value.path)
        result = []

        for dimension, job_chain in self.job_chains.items():
            thumbnail = DefaultThumbnailer(value.path, job_chain=job_chain)
            original = thumbnail.get_image(original_format)
            webp_thumbnail = thumbnail.get_image("WEBP")

            if dimension == 1:
                result = [
                    {"srcset": f"{webp_thumbnail.url} 1x", "type": webp_thumbnail.mimetype},
                    {"srcset": f"{original.url} 1x", "type": original.mimetype},
                ]

            else:
                result[0]["srcset"] = f"{result[0]['srcset']}, {webp_thumbnail.url} {dimension}x"
                result[1]["srcset"] = f"{result[1]['srcset']}, {original.url} {dimension}x"

        return result



from django.contrib.contenttypes.models import ContentType
from django.db.models.fields.files import FileField
from django.db.models.fields.files import FieldFile
from django.db.models.fields.files import ImageField
from django.db.models.fields.files import ImageFieldFile
from django.urls import reverse


class ProtectedFileMixin:
    @property
    def url(self):
        self._require_file()
        result_url = reverse(
            self.field.m_protected_view_name,
            kwargs={
                "contenttype_id": ContentType.objects.get_for_model(self.instance.__class__).id,
                "field": self.field.name,
                "pk": self.instance.pk,
            },
        )
        return result_url

    @property
    def original_url(self):
        self._require_file()
        return self.storage.url(self.name)


class ProtectedFieldFile(ProtectedFileMixin, FieldFile):
    pass


class ProtectedFileField(FileField):
    """
    A field that adds the mechanism of a protected file.
    """

    attr_class = ProtectedFieldFile

    def __init__(
        self,
        verbose_name=None,
        name=None,
        upload_to="protected",
        storage=None,
        protected_view_name="meringue-protected-file",
        **kwargs,
    ):
        self.m_protected_view_name = protected_view_name
        super().__init__(verbose_name, name, upload_to, storage, **kwargs)


class ProtectedImageFieldFile(ProtectedFileMixin, ImageFieldFile):
    pass


class ProtectedImageField(ImageField):
    """
    A field that adds the mechanism of a protected image.
    """

    attr_class = ProtectedImageFieldFile

    def __init__(
        self,
        verbose_name=None,
        name=None,
        width_field=None,
        height_field=None,
        protected_view_name="meringue-protected-file",
        **kwargs,
    ):
        kwargs.setdefault("upload_to", "protected")
        self.m_protected_view_name = protected_view_name
        super().__init__(verbose_name, name, width_field, height_field, **kwargs)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class Config(AppConfig):
    name = "meringue.protected"
    verbose_name = _("Protected")
    label = "meringue_protected"



import math

from PIL import Image

from meringue.thumbnail import constants
from meringue.thumbnail.exceptions import ActionError


def get_size(options, max_sizes=True):
    """
    Calc image size
    """

    new_size = options[constants.PROP_NEW_SIZE]
    if max_sizes:
        if options[constants.PROP_MAX_WIDTH] and new_size[0] > options[constants.PROP_MAX_WIDTH]:
            new_size[0] = options[constants.PROP_MAX_WIDTH]

        if options[constants.PROP_MAX_HEIGHT] and new_size[1] > options[constants.PROP_MAX_HEIGHT]:
            new_size[1] = options[constants.PROP_MAX_HEIGHT]

    return new_size


def crop(image, options):
    """
    Crop image
    """

    new_size = get_size(options)
    new_size = [int(new_size[0]), int(new_size[1])]

    current_size = options[constants.PROP_CURRENT_SIZE]
    thumb = Image.new(mode="RGBA", size=new_size, color=options[constants.PROP_BG_COLOR])

    # позиционируем новое изображение по X
    if options[constants.PROP_CROP_METHOD][0] == constants.CROP_METHOD_LEFT:
        x = 0
    elif options[constants.PROP_CROP_METHOD][0] == constants.CROP_METHOD_RIGTH:
        x = new_size[0] - current_size[0]
    else:
        # CROP_METHOD_CENTER
        x = (new_size[0] - current_size[0]) / 2

    # позиционируем новое изображение по Y
    if options[constants.PROP_CROP_METHOD][1] == constants.CROP_METHOD_TOP:
        y = 0
    elif options[constants.PROP_CROP_METHOD][1] == constants.CROP_METHOD_BOTTOM:
        y = new_size[1] - current_size[1]
    else:
        # CROP_METHOD_CENTER
        y = (new_size[1] - current_size[1]) / 2

    thumb.paste(image, (int(x), int(y)))
    image = thumb
    options[constants.PROP_CURRENT_SIZE] = new_size

    return image


def resize(image, options):
    """
    Resize image
    """

    new_size = get_size(options)
    current_size = options[constants.PROP_CURRENT_SIZE]
    method = options[constants.PROP_RESIZE_METHOD]
    strategy = options[constants.PROP_RESIZE_STRATEGY]

    if method in [constants.RESIZE_METHOD_COVER, constants.RESIZE_METHOD_CONTAIN]:
        # режим cover - заполнить всё новое изображение старым
        if method == constants.RESIZE_METHOD_COVER:
            res = max(
                new_size[0] / current_size[0],
                new_size[1] / current_size[1],
            )

        # режим contain - вписать изображение полностью в новый размер
        elif method == constants.RESIZE_METHOD_CONTAIN:
            res = min(
                new_size[0] / current_size[0],
                new_size[1] / current_size[1],
            )

        if strategy == constants.RESIZE_STRATEGY_DO_NOT_INCREASE_SIZE and res > 1:
            # отказываем в увеличении размера
            new_size = current_size.copy()

        elif strategy == constants.RESIZE_STRATEGY_DO_NOT_REDUCE_SIZE and res < 1:
            # отказываем в уменьшении размера
            new_size = current_size.copy()

        else:
            new_size = [math.floor(current_size[0] * res), math.floor(current_size[1] * res)]

    else:  # noqa: PLR5501
        # RESIZE_METHOD_STRETCH
        # при сжатии никак не меняем конечный размер

        if strategy != constants.RESIZE_STRATEGY_STANDART:
            msg = (
                f"The `{constants.RESIZE_METHOD_STRETCH}` method is not compatible "
                f"with the `{strategy}` strategy."
            )
            raise ActionError(msg)

    new_size = [int(new_size[0]), int(new_size[1])]

    if new_size[0] != current_size[0] or new_size[1] != current_size[1]:
        image = image.resize(new_size, resample=Image.LANCZOS)
        options[constants.PROP_CURRENT_SIZE] = new_size

    return image



from django.db.models import Q
from django.db.models.query import QuerySet
from django.utils import timezone


class SortingQuerySet(QuerySet):
    """
    QuerySet with methods for working with sorted positions
    """

    def correction_sorting(self):
        """
        This is a method to update/fix the sorting of the selected list of items.

        The sorting will be done according to the `queryset` sorting, so sorting can be controlled
        by executing `.order_by()` before calling the `correction_sorting` method.

        The selection for updating sorting can be pre-limited by filtering the list.
        """
        items = []
        sorting = 0
        for item in self:
            if item.sorting != sorting:
                item.sorting = sorting
                items.append(item)
            sorting += 1

        return self.bulk_update(items, ["sorting"])


class PublicationQuerySet(QuerySet):
    """
    QuerySet with methods for fetching published and unpublished elements.
    """

    def published(self, *args, **kwargs):
        """
        Method to getting published items.
        """
        from meringue.core.models import PublicationDatesMixin
        from meringue.core.models import PublicationMixin

        if issubclass(self.model, PublicationMixin):
            kwargs["is_published"] = True

        if issubclass(self.model, PublicationDatesMixin):
            now = timezone.now()
            args = (
                *args,
                Q(date_from__lte=now, date_to__gt=now)
                | Q(date_from__isnull=True, date_to__gt=now)
                | Q(date_from__lte=now, date_to__isnull=True)
                | Q(date_from__isnull=True, date_to__isnull=True),
            )

        return self.filter(*args, **kwargs)

    def unpublished(self, *args, **kwargs):
        """
        Method for getting unpublished items.
        """
        from meringue.core.models import PublicationDatesMixin
        from meringue.core.models import PublicationMixin

        if issubclass(self.model, PublicationMixin):
            kwargs["is_published"] = False

        if issubclass(self.model, PublicationDatesMixin):
            now = timezone.now()
            args = (*args, Q(date_from__gt=now) | Q(date_to__lte=now))

        return self.filter(*args, **kwargs)


class UniversalQuerySet(SortingQuerySet, PublicationQuerySet):
    """
    Generic QuerySet containing methods of all other QuerySets
    """



class ThumbnailError(Exception):
    """
    Base thumbnail exception
    """

    default_message = "Thumbnail Exception."
    default_code = "thumbnail_exception"

    def __init__(self, message: str | None = None, code: str | None = None):
        """
        Attributes:
            message: Custom error message.
            code: Custom error code.
        """
        self.message = message or self.default_message
        self.code = code or self.default_code
        super().__init__(f"Error `{self.code}`: {self.message}")


class WrongPropertyOptionError(ThumbnailError):
    """
    Exception wrong property option
    """

    default_message = "The value of `{option}` is invalid for property `{prop}`."
    default_code = "wrong_property_option"

    def __init__(self, prop: str, option: str, message: str | None = None, code: str | None = None):
        """
        Attributes:
            prop: The property with the problem.
            option: Value with which there was a problem.
            message: Custom error message.
            code: Custom error code.
        """
        message = (message or self.default_message).format(prop=prop, option=option)
        code = code or self.default_code
        super().__init__(message, code)


class ActionError(ThumbnailError):
    """
    Action exception
    """

    default_message = "Action exception."
    default_code = "action_exception"


class ThumbnailerError(ThumbnailError):
    """
    Thumbnailer exception
    """

    default_message = "Thumbnailer default exception."
    default_code = "thumbnailer_exception"


class WrongActionOrPropertyError(ThumbnailerError):
    """
    Wrong action or property exception
    """

    default_message = "Action or property `{job}` not registered."
    default_code = "wrong_action_or_property"

    def __init__(self, job: str, message: str | None = None, code: str | None = None):
        """
        Attributes:
            job: Action with problem.
            message: Custom error message.
            code: Custom error code.
        """
        message = (message or self.default_message).format(job=job)
        super().__init__(message, code)


class WrongFormatError(ThumbnailerError):
    """
    Wrong action or property exception
    """

    default_message = "`{out_format}` format is not supported."
    default_code = "wrong_action_or_property"

    def __init__(self, out_format: str, message: str | None = None, code: str | None = None):
        """
        Attributes:
            out_format: Problem format.
            message: Custom error message.
            code: Custom error code.
        """
        message = (message or self.default_message).format(out_format=out_format)
        super().__init__(message, code)



from drf_spectacular.views import SpectacularAPIView
from rest_framework.response import Response


__all__ = [
    "MeringueSpectacularAPIView",
]


class MeringueSpectacularAPIView(SpectacularAPIView):
    """
    Spectacular wrapper that patches the schema.
    """

    patcher = None

    def patch_schema(self, openapi_schema: dict):
        """
        Method patching OpenAPI3 object.

        Attributes:
            openapi_schema: OpenAPI Object.
        """

        if self.patcher:
            self.patcher.patch_schema(openapi_schema)

    def _get_schema_response(self, request):
        """
        Copy of [SpectacularAPIView._get_schema_response][drf_spectacular.views.SpectacularAPIView.
        _get_schema_response] but with schema patch
        """
        version = self.api_version or request.version or self._get_version_parameter(request)
        generator = self.generator_class(
            urlconf=self.urlconf, api_version=version, patterns=self.patterns
        )
        final_schema = generator.get_schema(request=request, public=self.serve_public)
        self.patch_schema(final_schema)
        return Response(
            data=final_schema,
            headers={
                "Content-Disposition": f'inline; filename="{self._get_filename(request, version)}"',
            },
        )



from pathlib import Path
from typing import Final
from urllib.parse import urljoin

from django.conf import settings


# CORE #############################################################################################

UPLOAD_RENAME_HANDLER: Final[str] = "meringue.core.upload_handlers.rename_handler"
"""
Path to method for renaming images on upload
"""

COP_YEAR: Final[int] = None
"""
Project start year for the copyright tag
"""

COP_YEARS_DIFF: Final[int] = 10
"""
Difference in years for which it is necessary to display the range of years

Specify 1 to display the period for the second year.
"""

CRYPTO_KEY: Final[str] = settings.SECRET_KEY[:32]
"""
Encryption key
"""

FRONTEND_URLS: Final[dict] = None
"""
A dict of links to the frontend
"""

FRONTEND_DOMAIN: Final[str] = None
"""
Domain for generating absolute links
"""


# API ##############################################################################################

API_ENABLE_ROOT_VIEW: Final[bool] = settings.DEBUG
"""
Option to enable or disable the root view of the [Router][meringue.api.routers.MeringueRouter]
"""


# THUMBNAIL ########################################################################################

THUMBNAIL_GENERATOR_CLASS: Final[str] = "meringue.thumbnail.generators.Thumbnailer"
"""
Thumbnail generator class.
"""

THUMBNAIL_STORAGE_GETTER: Final[str] = "meringue.thumbnail.storage.get_storage"
"""
Dotted path to a method that returns a store.
"""

THUMBNAIL_IMAGE_CLASS: Final[str] = "meringue.thumbnail.images.ThumbnailImage"
"""
Thumbnail image class.
"""

THUMBNAIL_DIR: Final[Path] = Path(settings.MEDIA_ROOT) / Path("m/thumbnail")
"""
Directory for saving thhumbnails.
"""

THUMBNAIL_URL: Final[str] = urljoin(settings.MEDIA_URL, "m/thumbnail")
"""
Url where thumbnails will be available.
"""

THUMBNAIL_PROPERTIES: Final[dict[str, str]] = {
    "cm": "meringue.thumbnail.properties.set_crop_method",
    "rm": "meringue.thumbnail.properties.set_resize_method",
    "rs": "meringue.thumbnail.properties.set_resize_strategy",
    "s": "meringue.thumbnail.properties.set_size",
    "maxw": "meringue.thumbnail.properties.set_max_width",
    "maxh": "meringue.thumbnail.properties.set_max_height",
    "c": "meringue.thumbnail.properties.set_bg_color",
}
THUMBNAIL_ACTIONS: Final[dict[str, str]] = {
    "crop": "meringue.thumbnail.actions.crop",
    "resize": "meringue.thumbnail.actions.resize",
}

THUMBNAIL_DEFAULT_CROP_METHOD: Final[list[str]] = ["center", "center"]
THUMBNAIL_DEFAULT_RESIZE_METHOD: Final[str] = "contain"
THUMBNAIL_DEFAULT_RESIZE_STRATEGY: Final[str] = "standart"
THUMBNAIL_DEFAULT_BG_COLOR: Final[tuple[int]] = (200, 200, 200, 0)


THUMBNAIL_SAVE_PARAMS_BY_FORMAT: Final[dict[str, dict]] = {
    "GIF": {"optimize": True},
    "JPEG": {"quality": 85, "optimize": True},
    "PNG": {"optimize": True, "compress_level": 5},
    "TIFF": {"quality": 85},
    "WEBP": {"quality": 85},
}
"""
List of default options for saving thumbnails images by format.
"""

# THUMBNAIL_IMAGE_OPTIMIZE_HANDLER: Final[str] = None

THUMBNAIL_DUMMYIMAGE_TEMPLATE: Final[str] = "//dummyimage.com/{width}x{height}/9e9e9e/424242.png"

THUMBNAIL_DEFAULT_FORMAT: Final[str] = "PNG"
"""
Default thumbnail image format.
"""


# PROTECTED ########################################################################################

PROTECTED_SERVE_WITH_NGINX: Final[bool] = not settings.DEBUG
"""
The option implies the distribution of protected files by nginx. Instead of serving the file in
response.

The view [protected_file_view][meringue.protected.views.protected_file_view] adds the X-Accel-Redirect header with a link to the file.
"""



import hashlib
import json
import mimetypes
from io import BytesIO
from pathlib import Path
from typing import Final

from django.core.files.storage import Storage
from django.db.models.utils import AltersData

from PIL import Image

from meringue.conf import m_settings
from meringue.thumbnail import constants
from meringue.thumbnail.storage import default_storage
from meringue.thumbnail.types import JobChainType


class ThumbnailImage(AltersData):
    """
    Thumbnail Image class.
    """

    def __init__(
        self,
        image_path: Path,
        job_chain: JobChainType,  # str  = "extra@1",
        out_format: str,
        image: Image,
        storage: Storage | None = None,
    ):
        """
        Attributes:
            image_path: Source image path.
            job_chain: Thumbnail job chain.
            out_format: Output image format.
            image: Thumbnail PIL image.
            storage: File storage.
        """

        self.source_image_path = image_path
        self.job_chain = job_chain
        self.out_format = out_format
        self.image = image
        self.storage = storage or default_storage
        self._saved = None

    @property
    def name(self) -> Path:
        """
        Relative path to thumbnail.

        We split the path into separate directories so as not to dump everything in one directory,
        which can be a problem with a large count of images.

        Returns:
            Thumbnail path.
        """

        filename = self.filename
        path = Path(filename[:2], filename[2:4], filename)
        return path

    @property
    def thumbnail_hash(self) -> str:
        """
        Calculates the thumbnail hash based on the job chain and source path.

        Returns:
            Thumbnail hash.
        """

        if getattr(self, "_thumbnail_hash", None) is None:
            params = [
                self.source_image_path.as_posix(),
                json.dumps(self.job_chain),
            ]
            salt = "|".join(params)
            self._thumbnail_hash = hashlib.md5(salt.encode()).hexdigest()  # noqa: S324

        return self._thumbnail_hash

    @property
    def file_extension(self) -> str:
        """
        Returns the file extension for the thumbnail.

        Returns:
            Thumbnail file extension.
        """

        return constants.EXTENSIONS_BY_FORMATS[self.out_format]

    @property
    def filename(self) -> str:
        """
        Returns the file name for the thumbnail.

        Returns:
            Thumbnail file name.
        """

        return f"{self.thumbnail_hash}{self.file_extension}"

    @property
    def path(self) -> Path:
        """Absolute path to thumbnail."""
        return self.storage.path(self.name)

    @property
    def url(self) -> str:
        """Url to thumbnail."""
        return self.storage.url(self.name)

    @property
    def is_supports_alpha(self) -> bool:
        """Returns the alpha channel support flag."""
        return self.out_format in constants.FORMATS_WITH_ALPHA_SUPPORT

    @property
    def mimetype(self) -> str:
        """
        Return thumbnail mimetype.

        Python mimetypes library doesn't know webp =(.

        Returns:
            Thumbnail mimetype.
        """

        if self.out_format == constants.FORMAT_WEBP:
            return "image/webp"

        guess_type = mimetypes.guess_type(str(self.absolute_path))[0]
        return guess_type[0]

    @property
    def saved(self) -> bool:
        """
        Check saved on disk thumbnail or not.

        Returns:
            Saved on disk flag.
        """

        if self._saved is None:
            self._saved = Path(self.path).exists()

        return self._saved

    def save(self, force=False, **kwargs):
        """
        Save image to disk.

        Attributes:
            **kwargs: Pillow imamge save params.
        """

        if (not force) and self.saved:
            return

        image = self.image.copy()

        if not self.is_supports_alpha:
            image = image.convert("RGB")

        params = {
            **m_settings.THUMBNAIL_SAVE_PARAMS_BY_FORMAT,
            **kwargs,
            "format": self.out_format,
        }

        tmp_image = BytesIO()
        image.save(tmp_image, **params)
        self.storage.save(self.name, tmp_image)
        self._saved = True

        # if m_settings.THUMBNAIL_IMAGE_OPTIMIZE_HANDLER:
        #     optimze_handler = import_string(m_settings.THUMBNAIL_IMAGE_OPTIMIZE_HANDLER)
        #     optimze_handler(self)

    save.alters_data = True


DefaultThumbnailImage: Final = m_settings.THUMBNAIL_IMAGE_CLASS



import logging
import re
from pathlib import Path

from django.conf import settings

from meringue.conf import m_settings
from meringue.thumbnail.generators import DefaultThumbnailer
from meringue.thumbnail.types import JobChainType


logger = logging.getLogger("meringue.thumbnail")


def _dummyimage(job_chain: str) -> str:
    """
    Return link to dummy image

    TODO:
     * configurated dummy image?

    Attributes:
        job_chain: Thumbnail job chain.

    Returns:
        Dummy image url.
    """

    if settings.DEBUG:
        size = (42, 42)
        job_chain.reverse()
        final_size = re.compile(r"s:(\d+)x(\d+)")
        for task in job_chain:
            try:
                size = final_size.search(task).groups()
                break
            except AttributeError:
                pass

        return m_settings.THUMBNAIL_DUMMYIMAGE_TEMPLATE.format(
            width=size[0],
            height=size[1],
        )

    return "%simages/noise.png" % settings.STATIC_URL
    # return "%simages/none.gif" % settings.STATIC_URL


def get_thumbnail(
    file_path: Path,
    job_chain: JobChainType,
    out_format: str = m_settings.THUMBNAIL_DEFAULT_FORMAT,
    **kwargs,
) -> str:
    """
    Shortcut to make single image preview

    Attributes:
        file_path: Source file path.
        job_chain: Thumbnail job chain.
        format: Output image format.
        **kwargs: Output image save extra params.

    Returns:
        Thumbnail url.
    """

    if not file_path.exists():
        msg = f"File `{file_path}` not found."
        logger.error(msg)
        return _dummyimage(job_chain)

    thumbnailer = DefaultThumbnailer(file_path, job_chain=job_chain)
    thumbnail_image = thumbnailer.get_thumbnail(out_format, **kwargs)
    return thumbnail_image.url



import json

from django.core.serializers.json import DjangoJSONEncoder


__all__ = [
    "OpenAPISchemaPatcher",
]


class OpenAPISchemaPatcher:
    """
    OpenApi3 patcher.

    https://swagger.io/specification/

    !!! info
        `json.loads` -> `json.dumps` - for render locale `__proxy__` (e.g. `gettext_lazy`)
    """

    def __init__(self):
        self._security_schemes = {}
        self._components_schemes = {}
        self._tags_schemes = {}

    def register_security_scheme(self, name: str, scheme: dict):
        """
        Register security scheme.

        Attributes:
            name: Security scheme name.
            scheme: Security scheme in [Security Scheme Object](https://swagger.io/specification/#s
                ecurity-scheme-object) or [Reference Object format](https://swagger.io/specificatio
                n/#reference-object).

        Raises:
            Exception: Security scheme already registered.
        """

        if name in self._security_schemes:
            msg = f"Security scheme named `{name}` is already registered"
            raise Exception(msg)

        self._security_schemes[name] = scheme

    def register_component_scheme(self, name: str, scheme: dict):
        """
        Register components scheme.

        Attributes:
            name: Component name.
            schema: [Formatted](https://swagger.io/specification/#schema-object) Component schema
                object.

        Raises:
            Exception: Scheme is already registered.
        """

        if name in self._components_schemes:
            msg = f"Component named `{name}` is already registered"
            raise Exception(msg)

        self._components_schemes[name] = scheme

    def register_tag(self, scheme: dict):
        """
        Register tag.

        Attributes:
            scheme: [Tag scheme object](https://swagger.io/specification/#tag-object).

        Raises:
            Exception: Tag is already registered.
        """
        name = scheme["name"]

        if name in self._tags_schemes:
            msg = f"A tag named `{name}` is already registered"
            raise Exception(msg)

        self._tags_schemes[name] = scheme

    def patch_description(self, openapi_schema: dict):
        """
        Adds a list of servers from the openapi schema source servers to the description.

        This method adds to the end of the description a list of servers that it takes from the
        [corresponding section](https://swagger.io/specification/#server-object) of the openapi
        schema.

        Attributes:
            openapi_schema: OpenAPI Object.
        """

        if not openapi_schema.get("servers", []):
            return

        openapi_schema.setdefault("info", {})
        openapi_schema["info"].setdefault("description", "")

        for server in openapi_schema["servers"]:
            name = server["description"]
            full_url = server["url"]
            clear_url = full_url.split("://")[1] if "://" in full_url else full_url
            if openapi_schema["info"]["description"]:
                openapi_schema["info"]["description"] += "\n\n"
            openapi_schema["info"]["description"] += f"{name} API: [{clear_url}]({full_url})"

    def patch_security_schemes(self, openapi_schema: dict):
        """
        Adds registered security schemes to the openapi schema.

        Attributes:
            openapi_schema: OpenAPI Object.

        Raises:
            Exception: Security scheme already registered.
        """

        if not self._security_schemes:
            return

        openapi_schema.setdefault("components", {})
        openapi_schema["components"].setdefault("securitySchemes", {})

        tmp_schemes = json.loads(json.dumps(self._security_schemes, cls=DjangoJSONEncoder))

        for name, schema in tmp_schemes.items():
            if name in openapi_schema["components"]["securitySchemes"]:
                msg = f"Security scheme named `{name}` exists in the original schema"
                raise Exception(msg)

            openapi_schema["components"]["securitySchemes"][name] = schema

    def patch_component_schemes(self, openapi_schema: dict):
        """
        Adds registered components schemas to the openapi schema.

        Attributes:
            openapi_schema: OpenAPI Object.

        Raises:
            Exception: Component schema is already registered.
        """

        if not self._components_schemes:
            return

        openapi_schema.setdefault("components", {})
        openapi_schema["components"].setdefault("schemas", {})

        tmp_components = json.loads(json.dumps(self._components_schemes, cls=DjangoJSONEncoder))

        for name, schema in tmp_components.items():
            if name in openapi_schema["components"]["schemas"]:
                msg = f"Component named `{name}` exists in the original schema"
                raise Exception(msg)

            openapi_schema["components"]["schemas"][name] = schema

    def patch_tags(self, openapi_schema: dict):
        """
        Adds registered components schemas to the openapi schema.

        Attributes:
            openapi_schema: OpenAPI Object.

        Raises:
            Exception: A tag exists in the original schema.
        """

        if not self._tags_schemes:
            return

        openapi_schema.setdefault("tags", [])
        exists_tags_schemes = (t["name"] for t in openapi_schema["tags"])
        tmp_tags = json.loads(json.dumps(self._tags_schemes, cls=DjangoJSONEncoder))

        for name, schema in tmp_tags.items():
            if name in exists_tags_schemes:
                msg = f"A tag named `{name}` exists in the original schema"
                raise Exception(msg)

            openapi_schema["tags"].append(schema)

    def patch_schema(self, openapi_schema: dict):
        """
        Fully patch the schema.
        """
        self.patch_description(openapi_schema)
        self.patch_security_schemes(openapi_schema)
        self.patch_component_schemes(openapi_schema)
        self.patch_tags(openapi_schema)
